"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const apputils_1 = require("@jupyterlab/apputils");
const notebook_1 = require("@jupyterlab/notebook");
const codemirror_1 = require("@jupyterlab/codemirror");
const fileeditor_1 = require("@jupyterlab/fileeditor");
const coreutils_1 = require("@jupyterlab/coreutils");
const docmanager_1 = require("@jupyterlab/docmanager");
const fileeditor_2 = require("@krassowski/jupyterlab_go_to_definition/lib/jumpers/fileeditor");
const notebook_2 = require("@krassowski/jupyterlab_go_to_definition/lib/jumpers/notebook");
require("codemirror/addon/hint/show-hint.css");
require("codemirror/addon/hint/show-hint");
require("../style/index.css");
require("lsp-editor-adapter/lib/codemirror-lsp.css");
const lsp_editor_adapter_1 = require("lsp-editor-adapter");
const completer_1 = require("@jupyterlab/completer");
const completion_1 = require("./completion");
const notebook_mapper_1 = require("./notebook_mapper");
const lsp_1 = require("./lsp");
const widgets_1 = require("@phosphor/widgets");
const rendermime_1 = require("@jupyterlab/rendermime");
const free_tooltip_1 = require("./free_tooltip");
const utils_1 = require("./utils");
const converter_1 = require("./converter");
/*
Feedback: anchor - not clear from docs
bundle - very not clear from the docs, interface or better docs would be nice to have
 */
// TODO: settings
const hover_modifier = 'Control';
class CodeMirrorAdapterExtension extends lsp_editor_adapter_1.CodeMirrorAdapter {
    constructor(connection, options, editor, create_tooltip) {
        super(connection, options, editor);
        this.marked_diagnostics = new Map();
        this.create_tooltip = create_tooltip;
        // @ts-ignore
        let listeners = this.editorListeners;
        let wrapper = this.editor.getWrapperElement();
        // detach the adapters contextmenu
        wrapper.removeEventListener('contextmenu', listeners.contextmenu);
        // show hover after pressing the modifier key
        wrapper.addEventListener('keydown', (event) => {
            if (!hover_modifier || utils_1.getModifierState(event, hover_modifier)) {
                this.show_next_tooltip = true;
                this.handleHover(this.last_hover_response);
            }
        });
        wrapper.addEventListener('keyup', () => {
            // TODO: the updates frequency and triggers will require a review and a clean up
            this.connection.sendChange();
        });
    }
    handleGoTo(locations) {
        this.remove_tooltip();
        // do NOT handle GoTo actions here
    }
    handleCompletion(completions) {
        // do NOT handle completion here
        // TODO: UNLESS the trigger character was typed!
    }
    ;
    static get_markup(response) {
        let contents = response.contents;
        // this causes the webpack to fail "Module not found: Error: Can't resolve 'net'" for some reason
        // if (lsProtocol.MarkedString.is(contents))
        ///  contents = [contents];
        if (typeof contents === "string")
            contents = [contents];
        if (!Array.isArray(contents))
            return contents;
        // now we have MarkedString
        let content = contents[0];
        if (typeof content === "string") {
            // coerce to MarkedString  object
            return {
                kind: 'plaintext',
                value: content
            };
        }
        else
            return {
                kind: "markdown",
                value: "```" + content.language + '\n' + content.value + "```"
            };
    }
    remove_tooltip() {
        // @ts-ignore
        this._removeHover(); // this removes underlines
        if (this._tooltip !== undefined)
            this._tooltip.dispose();
    }
    handleHover(response) {
        this.remove_tooltip();
        if (!response || !response.contents || (Array.isArray(response.contents) && response.contents.length === 0)) {
            return;
        }
        this.highlight_range(response.range, 'cm-lp-hover-available');
        this.last_hover_response = null;
        if (!this.show_next_tooltip) {
            this.last_hover_response = response;
            return;
        }
        const markup = CodeMirrorAdapterExtension.get_markup(response);
        // @ts-ignore
        let position = this.hoverCharacter;
        // TODO this is where the idea of mapping notebooks with an object pretending to be an editor has a weak side...
        let cm_editor = this.editor;
        if (cm_editor.get_editor_at !== undefined)
            cm_editor = cm_editor.get_editor_at(position);
        this._tooltip = this.create_tooltip(markup, cm_editor, position);
    }
    highlight_range(range, class_name) {
        // @ts-ignore
        let hover_character = this.hoverCharacter;
        let start;
        let end;
        if (range) {
            start = converter_1.PositionConverter.lsp_to_cm(range.start);
            end = converter_1.PositionConverter.lsp_to_cm(range.end);
        }
        else {
            // construct range manually using the token information
            let token = this.editor.getTokenAt(hover_character);
            start = { line: hover_character.line, ch: token.start };
            end = { line: hover_character.line, ch: token.end };
        }
        // @ts-ignore
        this.hoverMarker = this.editor.getDoc().markText(start, end, {
            className: class_name,
        });
    }
    handleMouseOver(event) {
        // proceed when no hover modifier or hover modifier pressed
        this.show_next_tooltip = !hover_modifier || utils_1.getModifierState(event, hover_modifier);
        return super.handleMouseOver(event);
    }
    handleDiagnostic(response) {
        /*
        TODO: the base class has the gutter support, like this
        this.editor.clearGutter('CodeMirror-lsp');
         */
        // Note: no deep equal for Sets or Maps in JS:
        // https://stackoverflow.com/a/29759699
        const markers_to_retain = new Set();
        // add new markers, keep track of the added ones
        let doc = this.editor.getDoc();
        let transform;
        let get_cell_id;
        const notebook_as_editor = this.editor;
        // duck typing: does it implement transform and get_cell_at?
        if (notebook_as_editor.transform !== undefined && notebook_as_editor.get_cell_at !== undefined) {
            transform = position => notebook_as_editor.transform(position);
            get_cell_id = position => notebook_as_editor.get_cell_at(position).id;
        }
        else {
            transform = position => position;
            get_cell_id = position => '';
        }
        response.diagnostics.forEach((diagnostic) => {
            const start = converter_1.PositionConverter.lsp_to_cm(diagnostic.range.start);
            const end = converter_1.PositionConverter.lsp_to_cm(diagnostic.range.end);
            const severity = lsp_1.diagnosticSeverityNames[diagnostic.severity];
            // what a pity there is no hash in the standard library...
            // we could use this: https://stackoverflow.com/a/7616484 though it may not be worth it:
            //   the stringified diagnostic objects are only about 100-200 JS characters anyway,
            //   depending on the message length; this could be reduced using some structure-aware
            //   stringifier; such a stringifier could also prevent the possibility of having a false
            //   negative due to a different ordering of keys
            // obviously, the hash would prevent recovery of info from the key.
            let diagnostic_hash = JSON.stringify(Object.assign({}, diagnostic, { 
                // the apparent marker position will change in the notebook with every line change for each marker
                // after the (inserted/removed) line - but such markers should not be invalidated,
                // i.e. the invalidation should be performed in the cell space, not in the notebook coordinate space,
                // thus we transform the coordinates and keep the cell id in the hash
                range: {
                    start: transform(start),
                    end: transform(end)
                }, cell: get_cell_id(start) }));
            markers_to_retain.add(diagnostic_hash);
            if (!this.marked_diagnostics.has(diagnostic_hash)) {
                let options = {
                    title: diagnostic.message + (diagnostic.source ? ' (' + diagnostic.source + ')' : ''),
                    className: 'cm-lsp-diagnostic cm-lsp-diagnostic-' + severity,
                };
                let marker;
                try {
                    marker = doc.markText(start, end, options);
                }
                catch (e) {
                    console.warn('Marking inspection (diagnostic text) failed, see following logs (2):');
                    console.log(diagnostic);
                    console.log(e);
                    return;
                }
                this.marked_diagnostics.set(diagnostic_hash, marker);
            }
            /*
            TODO and this:
              const childEl = document.createElement('div');
              childEl.classList.add('CodeMirror-lsp-guttermarker');
              childEl.title = diagnostic.message;
              this.editor.setGutterMarker(start.line, 'CodeMirror-lsp', childEl);
            do we want gutters?
           */
        });
        // remove the markers which were not included in the new message
        this.marked_diagnostics.forEach((marker, diagnostic_hash) => {
            if (!markers_to_retain.has(diagnostic_hash)) {
                this.marked_diagnostics.delete(diagnostic_hash);
                marker.clear();
            }
        });
    }
}
class NotebookAdapter {
    constructor(editor_widget, jumper, app, completion_manager, rendermime_registry) {
        this.widget = editor_widget;
        this.editor = editor_widget.content;
        this.completion_manager = completion_manager;
        this.jumper = jumper;
        this.rendermime_registry = rendermime_registry;
        this.init_once_ready().then();
    }
    is_ready() {
        return (this.widget.context.isReady &&
            this.widget.content.isVisible &&
            this.widget.content.widgets.length > 0 &&
            // @ts-ignore
            this.widget.model.metadata.get('language_info').name != '');
    }
    init_once_ready() {
        return __awaiter(this, void 0, void 0, function* () {
            let document_path = this.widget.context.path;
            console.log('LSP: waiting for', document_path, 'to fully load');
            yield utils_1.until_ready(this.is_ready.bind(this), -1);
            console.log('LSP:', document_path, 'ready for connection');
            let root_path = coreutils_1.PathExt.dirname(document_path);
            // TODO instead use mime types in servers.yml?
            // @ts-ignore
            let language = this.widget.model.metadata.get('language_info').name;
            // TODO
            // this.widget.context.pathChanged
            this.notebook_as_editor = new notebook_mapper_1.NotebookAsSingleEditor(this.widget, 
            // filter out line magics if the language is python
            language == 'python' ? '^%.*' : '', 
            // the pass is used to silence the linters in places where the line would otherwise be filtered out
            language == 'python' ? 'pass' : ''
            // TODO: ideally we would allow to substitute code for magics, etc
            );
            let cm_editor = this.notebook_as_editor;
            // TODO: reconsider where language, path and cwd belong
            // TODO: use native jupyterlab tooltips, fix autocompletion (see how it is done by default), using BOTH kernel and LSP
            //  change style for inspections so that unused variables are greyed out if there is enough info in diagnostics message
            console.log('LSP: will connect using root path:', root_path, 'and language:', language);
            this.connection = new lsp_editor_adapter_1.LspWsConnection({
                serverUri: 'ws://localhost/' + language,
                languageId: language,
                // paths handling needs testing on Windows and with other language servers
                // PathExt.join(root, jumper.cwd)
                // PathExt.join(root, jumper.path)
                rootUri: 'file:///' + root_path,
                documentUri: 'file:///' + document_path,
                documentText: this.get_notebook_content.bind(this),
            }).connect(new WebSocket('ws://localhost:3000/' + language));
            // @ts-ignore
            yield utils_1.until_ready(() => this.connection.isConnected, -1, 150);
            console.log('LSP:', document_path, 'connected');
            // @ts-ignore
            this.adapter = new CodeMirrorAdapterExtension(this.connection, {
                quickSuggestionsDelay: 50,
            }, cm_editor, (markup, cm_editor, position) => {
                const bundle = markup.kind === 'plaintext' ? { 'text/plain': markup.value } : { 'text/markdown': markup.value };
                const tooltip = new free_tooltip_1.FreeTooltip({
                    anchor: this.widget.content,
                    bundle: bundle,
                    editor: this.notebook_as_editor.cm_editor_to_ieditor.get(cm_editor),
                    rendermime: this.rendermime_registry,
                    position: converter_1.PositionConverter.cm_to_jl(this.notebook_as_editor.transform(position)),
                    moveToLineEnd: false
                });
                widgets_1.Widget.attach(tooltip, document.body);
                return tooltip;
            });
            // refresh server held state after every change
            // note this may be changed soon: https://github.com/jupyterlab/jupyterlab/issues/5382#issuecomment-515643504
            this.widget.model.contentChanged.connect(this.refresh_lsp_notebook_image.bind(this));
            // and refresh it after the cell was activated, just to make sure that the first experience is ok
            this.widget.content.activeCellChanged.connect(this.refresh_lsp_notebook_image.bind(this));
            // register completion connectors on cells
            // see https://github.com/jupyterlab/jupyterlab/blob/c0e9eb94668832d1208ad3b00a9791ef181eca4c/packages/completer-extension/src/index.ts#L198-L213
            const cell = this.widget.content.activeCell;
            const connector = new completion_1.LSPConnector({
                editor: cell.editor,
                connection: this.connection,
                coordinates_transform: (position) => this.notebook_as_editor.transform_to_notebook(cell, position)
            });
            const handler = this.completion_manager.register({
                connector,
                editor: cell.editor,
                parent: this.widget,
            });
            this.widget.content.activeCellChanged.connect((notebook, cell) => {
                const connector = new completion_1.LSPConnector({
                    editor: cell.editor,
                    connection: this.connection,
                    coordinates_transform: (position) => this.notebook_as_editor.transform_to_notebook(cell, position)
                });
                handler.editor = cell.editor;
                handler.connector = connector;
            });
        });
    }
    get_notebook_content() {
        return this.notebook_as_editor.getValue();
    }
    refresh_lsp_notebook_image(slot) {
        // TODO this is fired too often currently, debounce!
        this.connection.sendChange();
    }
}
class FileEditorAdapter {
    constructor(editor_widget, jumper, app, completion_manager, rendermime_registry) {
        this.widget = editor_widget;
        this.editor = editor_widget.content;
        this.rendermime_registry = rendermime_registry;
        this.app = app;
        // let root = PageConfig.getOption('serverRoot');
        let cm_editor = this.editor.editor;
        // TODO: reconsider where language, path and cwd belong
        let value = jumper.language;
        this.connection = new lsp_editor_adapter_1.LspWsConnection({
            serverUri: 'ws://localhost/' + value,
            languageId: value,
            // paths handling needs testing on Windows and with other language servers
            // PathExt.join(root, jumper.cwd)
            // PathExt.join(root, jumper.path)
            rootUri: 'file:///' + jumper.cwd,
            documentUri: 'file:///' + jumper.path,
            documentText: () => cm_editor.editor.getValue(),
        }).connect(new WebSocket('ws://localhost:3000/' + value));
        // @ts-ignore
        this.adapter = new CodeMirrorAdapterExtension(this.connection, {
            quickSuggestionsDelay: 50,
        }, cm_editor.editor, (markup, cm_editor, position) => {
            const bundle = markup.kind === 'plaintext' ? { 'text/plain': markup.value } : { 'text/markdown': markup.value };
            const tooltip = new free_tooltip_1.FreeTooltip({
                anchor: this.widget.content,
                bundle: bundle,
                editor: this.editor.editor,
                rendermime: rendermime_registry,
                position: converter_1.PositionConverter.cm_to_jl(position),
                moveToLineEnd: false
            });
            widgets_1.Widget.attach(tooltip, document.body);
            return tooltip;
        });
        // detach the adapters contextmenu for now:
        // @ts-ignore
        this.adapter.editor.getWrapperElement().removeEventListener('contextmenu', this.adapter.editorListeners.contextmenu);
        // TODO: actually we only need the connection... the tooltips and suggestions will need re-writing to JL standards anyway
        // @ts-ignore
        this.connection.on('goTo', (locations) => {
            // TODO: implement selector for multiple locations
            //  (like when there are multiple definitions or usages)
            let location = locations[0];
            // @ts-ignore
            let uri = location.uri;
            let current_uri = this.connection.getDocumentUri();
            // @ts-ignore
            let line = location.range.start.line;
            // @ts-ignore
            let column = location.range.start.character;
            if (uri == current_uri) {
                jumper.jump(jumper.getJumpPosition({ line: line, column: column }));
                return;
            }
            if (uri.startsWith('file://'))
                uri = uri.slice(7);
            console.log(uri);
            jumper.global_jump({
                // TODO: there are many files which are not symlinks
                uri: '.lsp_symlink/' + uri,
                editor_index: 0,
                line: line,
                column: column
            }, true);
        });
        const connector = new completion_1.LSPConnector({
            editor: this.editor.editor,
            connection: this.connection,
            coordinates_transform: null
        });
        completion_manager.register({
            connector,
            editor: this.editor.editor,
            parent: editor_widget,
        });
        console.log('Connected adapter');
    }
    get path() {
        return this.widget.context.path;
    }
    get_doc_position_from_context_menu() {
        // get the first node as it gives the most accurate approximation
        let leaf_node = this.app.contextMenuHitTest(() => true);
        let cm_editor = this.editor.editor;
        let { left, top } = leaf_node.getBoundingClientRect();
        // @ts-ignore
        let event = this.app._contextMenuEvent;
        // if possible, use more accurate position from the actual event
        // (but this relies on an undocumented and unstable feature)
        if (event !== undefined) {
            left = event.clientX;
            top = event.clientY;
            event.stopPropagation();
        }
        return cm_editor.editor.coordsChar({
            left: left,
            top: top,
        }, 'window');
    }
}
const file_editor_adapters = new Map();
/**
 * The plugin registration information.
 */
const plugin = {
    id: '@krassowski/jupyterlab-lsp:plugin',
    requires: [fileeditor_1.IEditorTracker, notebook_1.INotebookTracker, coreutils_1.ISettingRegistry, apputils_1.ICommandPalette, docmanager_1.IDocumentManager, completer_1.ICompletionManager, rendermime_1.IRenderMimeRegistry],
    activate: (app, fileEditorTracker, notebookTracker, settingRegistry, palette, documentManager, completion_manager, rendermime_registry) => {
        //CodeMirrorExtension.configure();
        fileEditorTracker.widgetUpdated.connect((sender, widget) => {
            console.log(sender);
            console.log(widget);
            // TODO?
            // adapter.remove();
            //connection.close();
        });
        fileEditorTracker.widgetAdded.connect((sender, widget) => {
            let fileEditor = widget.content;
            if (fileEditor.editor instanceof codemirror_1.CodeMirrorEditor) {
                let jumper = new fileeditor_2.FileEditorJumper(widget, documentManager);
                //let extension = new CodeMirrorExtension(fileEditor.editor, jumper);
                let adapter = new FileEditorAdapter(widget, jumper, app, completion_manager, rendermime_registry);
                file_editor_adapters.set(fileEditor.id, adapter);
                //extension.connect();
            }
        });
        let file_editor_commands = [
            {
                'id': 'lsp_get_definition',
                'execute': (connection, position) => connection.getDefinition(position),
                'isEnabled': (connection) => connection.isDefinitionSupported(),
                'label': 'Jump to definition',
            },
            {
                'id': 'lsp_get_type_definition',
                'execute': (connection, position) => connection.getTypeDefinition(position),
                'isEnabled': (connection) => connection.isTypeDefinitionSupported(),
                'label': 'Highlight type definition',
            },
            {
                'id': 'lsp_get_references',
                'execute': (connection, position) => connection.getReferences(position),
                'isEnabled': (connection) => connection.isReferencesSupported(),
                'label': 'Highlight references',
            }
        ];
        let is_context_menu_over_token = () => {
            let fileEditor = fileEditorTracker.currentWidget.content;
            let adapter = file_editor_adapters.get(fileEditor.id);
            let docPosition = adapter.get_doc_position_from_context_menu();
            if (!docPosition)
                return false;
            let ce_position = { line: docPosition.line, column: docPosition.ch };
            let token = adapter.editor.editor.getTokenForPosition(ce_position);
            return token.value !== '';
        };
        for (let cmd of file_editor_commands) {
            app.commands.addCommand(cmd.id, {
                execute: () => {
                    let fileEditor = fileEditorTracker.currentWidget.content;
                    let adapter = file_editor_adapters.get(fileEditor.id);
                    let docPosition = adapter.get_doc_position_from_context_menu();
                    cmd.execute(adapter.connection, docPosition);
                },
                isEnabled: () => {
                    let fileEditor = fileEditorTracker.currentWidget.content;
                    let adapter = file_editor_adapters.get(fileEditor.id);
                    return adapter && adapter.connection && cmd.isEnabled(adapter.connection);
                },
                isVisible: is_context_menu_over_token,
                label: cmd.label
            });
            app.contextMenu.addItem({
                selector: '.jp-FileEditor',
                command: cmd.id
            });
        }
        notebookTracker.widgetAdded.connect((sender, widget) => {
            // NOTE: assuming that the default cells content factory produces CodeMirror editors(!)
            let jumper = new notebook_2.NotebookJumper(widget, documentManager);
            new NotebookAdapter(widget, jumper, app, completion_manager, rendermime_registry);
        });
        function updateOptions(settings) {
            //let options = settings.composite;
            //Object.keys(options).forEach((key) => {
            //  if (key === 'modifier') {
            //    // let modifier = options[key] as KeyModifier;
            //    CodeMirrorExtension.modifierKey = modifier;
            //  }
            //});
        }
        settingRegistry
            .load(plugin.id)
            .then(settings => {
            updateOptions(settings);
            settings.changed.connect(() => {
                updateOptions(settings);
            });
        })
            .catch((reason) => {
            console.error(reason.message);
        });
        // Add an application command
        const cmdIds = {
            // in future add more commands
            jumpNotebook: 'lsp:notebook-jump',
            jumpFileEditor: 'lsp:file-editor-jump',
        };
        // Add the command to the palette.
        palette.addItem({ command: cmdIds.jumpNotebook, category: 'Notebook Cell Operations' });
        palette.addItem({ command: cmdIds.jumpFileEditor, category: 'Text Editor' });
        function isEnabled(tracker) {
            return () => tracker.currentWidget !== null
                &&
                    tracker.currentWidget === app.shell.currentWidget;
        }
        app.commands.addCommand(cmdIds.jumpNotebook, {
            label: 'Jump to definition',
            execute: () => {
                let notebook_widget = notebookTracker.currentWidget;
                let notebook = notebook_widget.content;
                let jumper = new notebook_2.NotebookJumper(notebook_widget, documentManager);
                let cell = notebook_widget.content.activeCell;
                let editor = cell.editor;
                let position = editor.getCursorPosition();
                let token = editor.getTokenForPosition(position);
                jumper.jump_to_definition({ token, origin: null }, notebook.activeCellIndex);
            },
            isEnabled: isEnabled(notebookTracker)
        });
        app.commands.addCommand(cmdIds.jumpFileEditor, {
            label: 'Jump to definition',
            execute: () => {
                let fileEditorWidget = fileEditorTracker.currentWidget;
                let fileEditor = fileEditorWidget.content;
                let jumper = new fileeditor_2.FileEditorJumper(fileEditorWidget, documentManager);
                let editor = fileEditor.editor;
                let position = editor.getCursorPosition();
                let token = editor.getTokenForPosition(position);
                jumper.jump_to_definition({ token, origin: null });
            },
            isEnabled: isEnabled(fileEditorTracker)
        });
        const bindings = [
            {
                selector: '.jp-Notebook.jp-mod-editMode',
                keys: ['Ctrl Alt B'],
                command: cmdIds.jumpNotebook
            },
            {
                selector: '.jp-FileEditor',
                keys: ['Ctrl Alt B'],
                command: cmdIds.jumpFileEditor
            },
        ];
        bindings.map(binding => app.commands.addKeyBinding(binding));
    },
    autoStart: true
};
/**
 * Export the plugin as default.
 */
exports.default = plugin;
